#pragma once

#include <Arduino.h>
#include <Wire.h>
#include "Config.h"
#include "Packet.h"


// Pin-Typen des Test-IC
// Muss identisch sein mit Enumeration IC_DB.PinType.
enum IcPinTypes_t {
   NcPin,
   GndPin,
   VccPin,
   InputPin,
   OutputPin,
   OpenCollectorPin,
   SdaPin,
   SclPin
};

// Steuerung der MCP23017.
class TestIcClass {
public:
   uint8_t numberOfPins; // Anzahl Pins des Test-IC.
   uint8_t sda = 0xFF; // I2C: SDA-Pin, 0xFF fr nicht gesetzt
   uint8_t scl = 0xFF; // I2C: SCL-Pin, 0xFF fr nicht gesetzt

   // Initialisiert die Schnittstelle zum Test-IC.
   // Muss vor jedem Test und am Testende aufgerufen werden.
   void initTest();

   bool setPinModes();  // Die IODIRx- und GPPUx-Register werden anhand des empfangenen Packets eingestellt.
   // Liefert false, falls unltiger Pin-Typ gesendet wurde.

   void setGpioData();  // Setzen der Output-Pins anhand des empfangenen Pakets.
   void buildGpioDataPacket(); // Liest die GPIOs ein und sendet Paket vom Typ 11.


   // Ermittelt die Nr des ZIF-Pins (nullbasiert, 0..27).
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   uint8_t getZifPin(uint8_t pinIndex);

   // Emittelt die I2C-Adresse des MCP23017, auf dem der Pin des Test-IC liegt.
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   uint8_t getMcpDeviceAddress(uint8_t pinIndex);

   // Emittelt die Register-Adresse der GPIO-Bank des MCP23017, auf der der Pin des Test-IC liegt.
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   uint8_t getGpioBankRegisterAddress(uint8_t pinIndex);

   // Emittelt die Register-Adresse der IODIR-Bank des MCP23017, auf der der Pin des Test-IC liegt.
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   uint8_t getIodirBankRegisterAddress(uint8_t pinIndex);

   // Emittelt die GPIO-Bank-Maske fr den Pin des Test-IC liegt.
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   uint8_t getBankPinMask(uint8_t pinIndex);

   // Sendet ein Paket ber die I2C-Schnittstelle
   // Output   
   // 0 ..success
   // 1 ..length to long for buffer
   // 2 ..address send, NACK received
   // 3 ..data send, NACK received
   // 4 ..other twi error(lost bus arbitration, bus error, ..)
   // 5 ..timeout
   uint8_t sendpacket(uint8_t* data, uint8_t len); // true, wenn erfolgreich versendet.
   // Empfngt ein Paket ber die I2C-Schnitstelle
   // Output   number of bytes read
   uint8_t readpacket(uint8_t deviceAdr, uint8_t* data, uint8_t len);


protected:
   void reset(); // Reset der MCP23017.

   // bertgt Value in das passende Bit den MCP-Puffer.
   // pinIndex nullbasierte Pin-Nummer des IC-Pins [0..(NumberOfPins - 1]).
   void setBufferPinValue(uint8_t pinIndex, uint8_t value);
   uint8_t getBufferPinValue(uint8_t pinIndex);

   void xmitPinModes(); // bertrgt den Puffer in die IODIRx-Register.
   void xmitPinValue(); // bertrgt den Puffer in die GPIOx-Register.
   void xmitPullUp();   // bertrgt den Puffer in die GPPUx-Register.
   void readAllPinValues(); // Liest die GPIOx-Register in den MCPBuffer;
};

