﻿using IC_DB;
using System.Reflection;

namespace IC_Tester_Controls
{
   /// <summary>
   /// Steuerelement zur Bearbeitung von <see cref="IcDefinition"/>-Objekten.
   /// </summary>
   public partial class IcDefinitionEditor : UserControl
   {
      /// <summary>
      /// Aufzählung der Aktionsarten.
      /// </summary>
      public enum IcEditorAction
      {
         /// <summary>Die Daten sollen gespeichert werden.</summary>
         Save,
         /// <summary>Änderungen sollen verworfen werden.</summary>
         Abort
      }

      /// <summary>
      /// Stellt die Methode dar, die das <see cref="ActionSelected"/>-Ereignis der <see cref="IcDefinitionEditor"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="IcEditorActionEventArgs"/>-Objekt, das die Ereignisdaten enthält.</param>
      public delegate void ActionEventHandler(object sender, IcEditorActionEventArgs e);
      /// <summary>
      /// Tritt ein, wenn die Bearbeitung abgeschlossen ist.
      /// </summary>
      public event ActionEventHandler? ActionSelected;


      private IcDefinition ic = new IcDefinition("");
      

      /// <summary><b>true</b>, wenn Daten geändert wurden.</summary>
      public bool dirty = false;

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="IcDefinitionEditor"/>-Klasse.
      /// </summary>
      public IcDefinitionEditor()
      {
         InitializeComponent();
      }

   /// <summary>
   /// Füllt die Steuerelemente mit den Daten dem übergebenen <see cref="IcDefinition"/>-Objekt.
   /// </summary>
   /// <param name="ic">Das <see cref="IcDefinition"/>-Objekt, das angezeigt werden soll.</param>
   public void fill(IcDefinition ic)
      {
         this.ic = ic;
         tbName.Text = ic.Name;
         tbDescription.Text = ic.Description;
         tbHints.Text = ic.Notice;
         nudPins.Value = ic.NumberOfPins;
         pbCircuit.Image = ic.CircuitDiagram;
         cmdSaveDef.Enabled = false;
         dirty = false;
      }

      /// <summary>
      /// Füllt die Steuerlemente mit einen neu angelegten <see cref="IcDefinition"/>-Objekt.
      /// </summary>
      public void clear() { fill(new IcDefinition("")); cmdSaveDef.Enabled = false; }



      // Reaktion auf das Anklicken der Schaltfläche "Speichern".
      // Übernimmt die Daten aus den Steuerelementen in das interne IcDefinition-Objekt und
      // löst das Ereignis <see cref="ActionSelected"/> aus.
      private void cmdSaveDef_Click(object sender, EventArgs e)
      {
         ic.Name = tbName.Text.Trim();
         ic.Description = tbDescription.Text.Trim();
         ic.Notice = tbHints.Text.Trim();
         ic.NumberOfPins = (int)nudPins.Value;
         ic.CircuitDiagram = (Bitmap)pbCircuit.Image;

         ActionSelected?.Invoke(this, new IcEditorActionEventArgs(IcEditorAction.Save, ic));
      }

      private void cmdAbort_Click(object sender, EventArgs e)
      {
         if (dirty)
         {
            var rtc = MessageBox.Show("Änderungern verwerfen?", "IC-Tester", MessageBoxButtons.OKCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2);
            if (rtc == DialogResult.OK) { ActionSelected?.Invoke(this, new IcEditorActionEventArgs(IcEditorAction.Abort, ic)); }
         }
         else
            ActionSelected?.Invoke(this, new IcEditorActionEventArgs(IcEditorAction.Abort, ic));
      }


      // Übernimmt ein per DragAndDrop übergebene Image-Datei in die PictureBox 'pbCircuit'.
      private void pbCircuit_DragDrop(object sender, DragEventArgs e)
      {
         if (e.Data is not null)
         {
            var data = e.Data.GetData(DataFormats.FileDrop);
            if (data != null)
            {
               var fileNames = data as string[];
               if (fileNames is not null && fileNames.Length > 0)
               {
                  pbCircuit.Image = Image.FromFile(fileNames[0]);
                  ItemChanged(this, new EventArgs());
               }
            }
         }
      }

      // Pin-Editor-Fenster aufrufen.
      private void cmdPins_Click(object sender, EventArgs e)
      {
         ic.Name = tbName.Text.Trim();
         ic.Description = tbDescription.Text.Trim();
         ic.Notice = tbHints.Text.Trim();
         ic.NumberOfPins = (int)nudPins.Value;
         ic.CircuitDiagram = (Bitmap)pbCircuit.Image;
         PinsEditor pe = new PinsEditor(ic);
         pe.ShowDialog();
         if (pe.Dirty) { ItemChanged(this, new EventArgs()); }
      }

      private void pbCircuit_Paint(object sender, PaintEventArgs e)
      {
         if (pbCircuit.Image is null)
         {
            if (pbCircuit.Enabled)
               e.Graphics.DrawString("Bild hierher ziehen\n350x200 Pixel²", this.Font, Brushes.Black, new Point(10, 10));
            else
               e.Graphics.DrawString("Bild hierher ziehen\n350x200 Pixel²", this.Font, Brushes.LightGray, new Point(10, 10));
         }
      }

      private void IcDefinitionEditor_Load(object sender, EventArgs e)
      {
         pbCircuit.AllowDrop = true;
      }

      private void pbCircuit_DragEnter(object sender, DragEventArgs e)
      {
         e.Effect = DragDropEffects.Copy;
      }

      private void ItemChanged(object sender, EventArgs e)
      {
         cmdSaveDef.Enabled = true;
         dirty = true;
      }


      // Test-Editor-Fenster aufrufen.
      private void cmdEditTest_Click(object sender, EventArgs e)
      {
         ic.Name = tbName.Text.Trim();
         ic.Description = tbDescription.Text.Trim();
         ic.Notice = tbHints.Text.Trim();
         ic.NumberOfPins = (int)nudPins.Value;
         ic.CircuitDiagram = (Bitmap)pbCircuit.Image;
         TestEditor te = new TestEditor(ic);
         te.ShowDialog();
         if (te.IsDirty) { 
            ItemChanged(this, new EventArgs()); 
         }
      }
   }

   /// <summary>
   /// Argumente für das Ereignis <see cref="IcDefinitionEditor.ActionEventHandler"/>.
   /// </summary>
   public class IcEditorActionEventArgs
   {
      /// <summary>Art der Aktion.</summary>
      public IcDefinitionEditor.IcEditorAction Action;
      /// <summary>Das zugehörige <see cref="IcDefinition"/>-Objekt.</summary>
      public IcDefinition? Ic;


      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="IcEditorActionEventArgs"/>-Klasse.
      /// </summary>
      /// <param name="action">Die Aktion.</param>
      /// <param name="ic">Das zugehörige <see cref="IcDefinition"/>-Objekt.</param>
      public IcEditorActionEventArgs(IcDefinitionEditor.IcEditorAction action, IcDefinition? ic)
      {
         Action = action;
         Ic = ic;
      }
   }
}
