﻿using System.Diagnostics;
using System.Runtime.InteropServices;


namespace IC_Tester_Controls
{
   /// <summary>
   /// Installliert einen Hook auf das MouseDown-Event.
   /// </summary>
   public static class MouseHook
   {
      /// <summary>
      /// Maus-Ereginisdaten
      /// </summary>
      public class MousHookEventArgs
      {
         /// <summary>Maus-Postion in Screen-Koordinaten.</summary>
         public int X;
         /// <summary>Maus-Postion in Screen-Koordinaten.</summary>
         public int Y;

         internal MousHookEventArgs(int x, int y)
         {
            X = x;
            Y = y;
         }
      }
      /// <summary>
      /// Tritt beim Drücken der linken Maustaste ein, auch außerhalb des Controls / Forms.
      /// </summary>
      /// <param name="sender">Quelle des Ereignisses. Ist <b>bull</b>.</param>
      /// <param name="e">Mouse </param>
      public delegate void MouseActionEventHandler(object? sender, MousHookEventArgs e);

      /// <summary>Tritt ein, wenn ein Item des Menüs zur Auswahl der Bearbeitung angklickt wurde.</summary>
      public static event MouseActionEventHandler? MouseDown;

      /// <summary>
      /// Startet den Hook.
      /// </summary>
      public static void Start() => _hookID = SetHook(_proc);
      /// <summary>
      /// Stoppt den Hook.
      /// </summary>
      public static void Stop() => UnhookWindowsHookEx(_hookID);

      private static LowLevelMouseProc _proc = HookCallback;
      private static IntPtr _hookID = IntPtr.Zero;

      private static IntPtr SetHook(LowLevelMouseProc proc)
      {
         using (Process curProcess = Process.GetCurrentProcess())
         using (ProcessModule? curModule = curProcess.MainModule)
         {
            if (curModule == null)
               throw new InvalidOperationException("Cannot get Main module");
            else
            {
               var moduleName = curModule.ModuleName;
               if (moduleName == null)
                  throw new InvalidOperationException("Cannot get modulke name");
               else
                  return SetWindowsHookEx(WH_MOUSE_LL, proc, GetModuleHandle(moduleName), 0);
            }
         }
      }

      private delegate IntPtr LowLevelMouseProc(int nCode, IntPtr wParam, IntPtr lParam);

      private static IntPtr HookCallback(int nCode, IntPtr wParam, IntPtr lParam)
      {
         if (nCode >= 0 && MouseMessages.WM_LBUTTONDOWN == (MouseMessages)wParam)
         {
            var hs = Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
            if (hs != null)
            {
               MSLLHOOKSTRUCT hookStruct = (MSLLHOOKSTRUCT)hs;
               MouseDown?.Invoke(null, new MousHookEventArgs(hookStruct.pt.x, hookStruct.pt.y));
            }
         }
         return CallNextHookEx(_hookID, nCode, wParam, lParam);
      }

      private const int WH_MOUSE_LL = 14;

      private enum MouseMessages
      {
         WM_LBUTTONDOWN = 0x0201,
         WM_LBUTTONUP = 0x0202,
         WM_MOUSEMOVE = 0x0200,
         WM_MOUSEWHEEL = 0x020A,
         WM_RBUTTONDOWN = 0x0204,
         WM_RBUTTONUP = 0x0205
      }

      [StructLayout(LayoutKind.Sequential)]
      private struct POINT
      {
         public int x;
         public int y;
      }

      [StructLayout(LayoutKind.Sequential)]
      private struct MSLLHOOKSTRUCT
      {
         public POINT pt;
         public uint mouseData, flags, time;
         public IntPtr dwExtraInfo;
      }

      [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
      private static extern IntPtr SetWindowsHookEx(int idHook,
        LowLevelMouseProc lpfn, IntPtr hMod, uint dwThreadId);

      [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
      [return: MarshalAs(UnmanagedType.Bool)]
      private static extern bool UnhookWindowsHookEx(IntPtr hhk);

      [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
      private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode,
        IntPtr wParam, IntPtr lParam);

      [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
      private static extern IntPtr GetModuleHandle(string lpModuleName);
   }
}
