﻿using System.Drawing.Design;
using System.ComponentModel;

namespace IC_Tester_Controls
{
   /// <summary>
   /// <img alt="" src="\media\VerticalLabel.png"></img> Ein Bezeichner-Steuerlement mit vertikaler Schriftrichtung.
   /// </summary>
   [DefaultProperty("Text")]
   [DefaultEvent("Click")]
   public partial class PinNameLabel : UserControl
   {
      private Bitmap bmp = new Bitmap(1, 1);

      /// <summary>
      /// Ausrichtung des Schriftzuges
      /// </summary>
      public enum DrawMode
      {
         /// <summary>Die Schrift wird von unten nach oben ausgegeben.</summary>
         BottomUp,
         /// <summary>Die Schrift wird von oben nach unten ausgegeben.</summary>
         TopDown
      }

      /// <summary>
      /// Stellt die Methode dar, die das <see cref="NameChanged"/>-Ereignis der <see cref="PinNameLabel"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="EventArgs"/>-Objekt.</param>
      public delegate void NameChangedHandler(object sender, EventArgs e);
      /// <summary>
      /// Tritt ein, wenn der Text geändert wurde.
      /// </summary>
      public  event NameChangedHandler? NameChanged;


      /// <summary>
      /// Initialisiert eine neue Instant der <see cref="PinNameLabel"/>-Klasse.
      /// </summary>
      public PinNameLabel()
      {
         // Dieser Aufruf ist für den Designer erforderlich.
         InitializeComponent();
      }

      #region  OnPaint 
      // Der Text wird mit einem Standard-Label-Control generiert.
      // Die Ausgabe erfolgt dabei in eine Bitmap, diese wird entsprechend gedreht.
      private void CreateImage()
      {
         if (lblDummy is null)
         {
            bmp = new Bitmap(1, 1);
            return;
         }

         lblDummy.AutoSize = mAutoSize;
         lblDummy.Width = Height;
         lblDummy.Height = Width;

         lblDummy.TextAlign = RotateAlignment(mTextDrawMode, mTextAlign);

         if (mTextDrawMode == DrawMode.BottomUp)
         {
            lblDummy.Padding = new Padding(Padding.Bottom, Padding.Left, Padding.Top, Padding.Right); // Rotieren des Padding
         }
         else
         {
            lblDummy.Padding = new Padding(Padding.Top, Padding.Right, Padding.Bottom, Padding.Left);
         }


         lblDummy.Font = Font;
         lblDummy.AutoEllipsis = mAutoEllipsis;

         lblDummy.ForeColor = ForeColor;
         lblDummy.BackColor = BackColor;


         Image? tmpImage = null;
         if (mImage is not null)
         {
            tmpImage = (Image)mImage.Clone();
            if (mTextDrawMode == DrawMode.BottomUp)
            {
               tmpImage.RotateFlip(RotateFlipType.Rotate90FlipNone);
            }
            else
            {
               tmpImage.RotateFlip(RotateFlipType.Rotate270FlipNone);
            }
         }
         lblDummy.Image = tmpImage;
         lblDummy.ImageAlign = RotateAlignment(mTextDrawMode, mImageAlign);

         lblDummy.Text = Text;
         lblDummy.Refresh();

         if (lblDummy.Width == 0)
         {
            bmp = new Bitmap(1, 1);
            return;
         }

         bmp = new Bitmap(lblDummy.Width, lblDummy.Height);
         lblDummy.DrawToBitmap(bmp, new Rectangle(0, 0, bmp.Width, bmp.Height));

         if (mTextDrawMode == DrawMode.BottomUp)
         {
            bmp.RotateFlip(RotateFlipType.Rotate270FlipNone);
         }
         else
         {
            bmp.RotateFlip(RotateFlipType.Rotate90FlipNone);
         }

         if (AutoSize)
         {
            Width = bmp.Width;
            Height = bmp.Height;
         }
      }

      /// <summary>
      /// Rotiert ContentAlignment-Object um 90° im oder gegen den Urzeigersinn.
      /// </summary>
      /// <param name="DrawMode">Drehrichtung.</param>
      /// <param name="Align">Zu rotierender Ausgangswert.</param>
      /// <returns>Rotierter Wert.</returns>
      private ContentAlignment RotateAlignment(DrawMode DrawMode, ContentAlignment Align)
      {
         if (DrawMode == DrawMode.BottomUp)
         {
            switch (Align)
            {
               case var @case when @case == ContentAlignment.BottomCenter:
                  {
                     return ContentAlignment.MiddleLeft;
                  }
               case var case1 when case1 == ContentAlignment.BottomLeft:
                  {
                     return ContentAlignment.TopLeft;
                  }
               case var case2 when case2 == ContentAlignment.BottomRight:
                  {
                     return ContentAlignment.BottomLeft;
                  }
               case var case3 when case3 == ContentAlignment.MiddleCenter:
                  {
                     return ContentAlignment.MiddleCenter;
                  }
               case var case4 when case4 == ContentAlignment.MiddleLeft:
                  {
                     return ContentAlignment.TopCenter;
                  }
               case var case5 when case5 == ContentAlignment.MiddleRight:
                  {
                     return ContentAlignment.BottomCenter;
                  }
               case var case6 when case6 == ContentAlignment.TopCenter:
                  {
                     return ContentAlignment.MiddleRight;
                  }
               case var case7 when case7 == ContentAlignment.TopLeft:
                  {
                     return ContentAlignment.TopRight;
                  }
               case var case8 when case8 == ContentAlignment.TopRight:
                  {
                     return ContentAlignment.BottomRight;
                  }
            }
         }
         else
         {
            switch (mTextAlign)
            {
               case var case9 when case9 == ContentAlignment.BottomCenter:
                  {
                     return ContentAlignment.MiddleRight;
                  }
               case var case10 when case10 == ContentAlignment.BottomLeft:
                  {
                     return ContentAlignment.BottomRight;
                  }
               case var case11 when case11 == ContentAlignment.BottomRight:
                  {
                     return ContentAlignment.TopRight;
                  }
               case var case12 when case12 == ContentAlignment.MiddleCenter:
                  {
                     return ContentAlignment.MiddleCenter;
                  }
               case var case13 when case13 == ContentAlignment.MiddleLeft:
                  {
                     return ContentAlignment.BottomCenter;
                  }
               case var case14 when case14 == ContentAlignment.MiddleRight:
                  {
                     return ContentAlignment.TopCenter;
                  }
               case var case15 when case15 == ContentAlignment.TopCenter:
                  {
                     return ContentAlignment.MiddleLeft;
                  }
               case var case16 when case16 == ContentAlignment.TopLeft:
                  {
                     return ContentAlignment.BottomLeft;
                  }
               case var case17 when case17 == ContentAlignment.TopRight:
                  {
                     return ContentAlignment.TopLeft;
                  }
            }
         }
         throw new ArgumentException("RotateAlignment: Die Wertekomination ist unzulässig. DrawMode = " + DrawMode.ToString() + ", Align: " + Align.ToString());
      }

      /// <summary>
      /// Zeichnet das Steuerelement.
      /// </summary>
      /// <param name="e">Ein PaintEventArgs, das die Ereignisdaten enthält.</param>
      protected override void OnPaint(PaintEventArgs e)
      {
         e.Graphics.Clear(BackColor);
         CreateImage();
         e.Graphics.DrawImage(bmp, new Point(0, 0));
         base.OnPaint(e);
      }
      #endregion

      #region  Properties 
      #region  AutoEllipsis 
      private bool mAutoEllipsis = false;
      /// <summary>
      /// Ruft einen Wert ab oder legt einen Wert fest, der angibt, ob das Auslassungszeichen (...) auf der rechten Seite von Label angezeigt wird. Das Auslassungszeichen weist darauf hin, dass der Label-Text über die angegebene Länge von Label hinausgeht.
      /// </summary>
      /// <value><b>True</b>, wenn der zusätzliche Bezeichnungstext durch Auslassungszeichen gekennzeichnet werden soll, andernfalls <b>False</b>. Die Standardeinstellung ist <b>False</b>.</value>
      [DefaultValue(false)]
      [Description("Aktiviert die automatische Behandlung von Text, der über die Höhe des Steuerelements hinaus geht.")]
      public bool AutoEllipsis
      {
         get
         {
            return mAutoEllipsis;
         }
         set
         {
            mAutoEllipsis = value;
            Refresh();
         }
      }
      #endregion
      #region  AutoSize 
      private bool mAutoSize = true;
      /// <summary>
      /// Ruft einen Wert ab, der angibt, ob die Größe des Steuerelements automatisch an dessen Inhalt angepasst wird, oder legt diesen fest.
      /// </summary>
      /// <value><b>True</b>, wenn das Steuerelement seine Breite dem Inhalt möglichst genau anpasst, andernfalls <b>False</b>. Der Standardwert ist <b>True</b>.</value>
      /// <remarks>Dies ist etwas anders, als das Verhalten des Steuerelements <see cref="Label.AutoSize">Label</see>.</remarks>
      [DefaultValue(true)]
      [Description("Aktiviert die automatische Größenanpassung.")]
      public override bool AutoSize
      {
         get
         {
            return mAutoSize;
         }
         set
         {
            mAutoSize = value;
            Refresh();
         }
      }
      #endregion
      #region  Image und zugehöriges 
      private Image? mImage = null;
      /// <summary>
      /// Ruft das im <see cref="PinNameLabel"/> dargestellte Bild ab oder legt dieses fest.
      /// </summary>
      /// <value>Ein Image, das im <see cref="PinNameLabel"/> dargestellt wird. Die Standardeinstellung ist <b>Nothing</b>.</value>
      /// <remarks>Diese Eigenschaft entspricht der Eigenschaft <see cref="Label.Image"/>. Das Bild wird nicht mit rotiert!</remarks>
      [Description("Das Bild, das auf dem Steuerelement angezeigt wird.")]
      public Image? Image
      {
         get
         {
            return mImage;
         }
         set
         {
            mImage = value;
            Refresh();
         }
      }

      private ContentAlignment mImageAlign = ContentAlignment.MiddleCenter;
      /// <summary>
      /// Ruft die Ausrichtung eines Bilds auf dem Steuerelement ab oder legt diese fest.
      /// </summary>
      /// <value>Ein Image, das im <see cref="PinNameLabel"/> dargestellt wird. Die Standardeinstellung ist <b>Nothing</b>.</value>
      /// <remarks>Diese Eigenschaft entspricht der Eigenschaft <see cref="Label.ImageAlign"/>.</remarks>
      [DefaultValue(ContentAlignment.MiddleCenter)]
      [Description("Die Ausrichtung des Bildes, das auf dem Steuerelement angezeigt wird.")]
      public ContentAlignment ImageAlign
      {
         get
         {
            return mImageAlign;
         }
         set
         {
            mImageAlign = value;
            Refresh();
         }
      }
      #endregion
      #region  Padding 
      /// <summary>
      /// Ruft die Innenabstände des Controls ab oder legt diese fest.
      /// </summary>
      /// <value>Ein <see cref="Padding"/>, das die internen Abstandsmerkmale des Steuerelements darstellt.</value>
      /// <remarks>Diese Eigenschaft entspricht der Eigenschaft <see cref="Control.Padding"/>.</remarks>
      [DefaultValue(typeof(Padding), "0, 0, 0, 0")]
      [Description("Gibt die Innenabstände des Controls an.")]
      public new Padding Padding
      {
         get
         {
            return base.Padding;
         }
         set
         {
            base.Padding = value;
            Refresh();
         }
      }
      #endregion
      #region  Text 
      /// <summary>
      /// Ruft den diesem Steuerelement zugeordneten Text ab oder legt diesen fest.
      /// </summary>
      /// <value>Der diesem Steuerelement zugeordnete Text.</value>
      /// <remarks>Diese Eigenschaft entspricht der Eigenschaft <see cref="Label.Text"/>.</remarks>
      [Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
      [SettingsBindable(true)]
      [Browsable(true)]
      [Bindable(true)]
      [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
      [EditorBrowsable(EditorBrowsableState.Always)]
      [Description("Der dem Steuerelement zugeordnete Text.")]
      public override string Text
      {
         get
         {
            return base.Text;
         }
         set
         {
            base.Text = value;
            Refresh();
         }
      }
      #endregion
      #region  TextAlign 
      private ContentAlignment mTextAlign = ContentAlignment.TopLeft;
      /// <summary>
      /// Ruft die Textausrichtung in der Bezeichnung ab oder legt diese fest.
      /// </summary>
      /// <value>Einer der <see cref="ContentAlignment"/>-Werte. Die Standardeinstellung ist <see cref="ContentAlignment.TopLeft">TopLeft</see>.</value>
      /// <remarks>Diese Eigenschaft entspricht der Eigenschaft <see cref="Label.TextAlign"/>.</remarks>
      [DefaultValue(ContentAlignment.TopLeft)]
      [Description("Bestimmt die Position innerhalb des Steuerelements.")]
      public virtual ContentAlignment TextAlign
      {
         get
         {
            return mTextAlign;
         }
         set
         {
            mTextAlign = value;
            Refresh();
         }
      }
      #endregion
      #region  TextDrawMode 
      private DrawMode mTextDrawMode = DrawMode.BottomUp;
      /// <summary>
      /// Richtung des Schriftzuges.
      /// </summary>
      /// <value>Richtung des Schriftzuges. Die Standardeinstellung ist <see cref="DrawMode.BottomUp"/>.</value>
      [DefaultValue(DrawMode.BottomUp)]
      [Description("Legt die Schriftrichtung fest.")]
      public DrawMode TextDrawMode
      {
         get
         {
            return mTextDrawMode;
         }
         set
         {
            mTextDrawMode = value;
            Refresh();
         }
      }
      #endregion
      #region  Hidden Property RightToLeft 
      /// <summary>
      /// -
      /// </summary>
      /// <value>-</value>
      [Browsable(false)]
      [Bindable(false)]
      [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
      [EditorBrowsable(EditorBrowsableState.Never)]
      [Obsolete]
#pragma warning disable CS0809
      public override RightToLeft RightToLeft { get; set; }
#pragma warning restore CS0809
      // Warnung 'Der veraltete Member "VerticalLabel.RightToLeft" überschreibt den nicht veralteten Member "Control.RightToLeft".'
      // kan ignoriert werden.
      #endregion
      #endregion

      // Sorgt für das Neuzeichnen.

      private void VerticalLabel_Resize(object sender, EventArgs e)
      {
         Refresh();
      }

     
      private void VerticalLabel_MouseClick(object sender, MouseEventArgs e)
      {
         Color ColorSave;
         ColorSave = BackColor;
         BackColor = Color.LightBlue;
         PinNameEditor pinNameEditor = new PinNameEditor(Text);
         pinNameEditor.Location = this.PointToScreen(Point.Empty);
         pinNameEditor.ShowDialog();
         if (Text != pinNameEditor.PinName)
         {
            Text = pinNameEditor.PinName;
            NameChanged?.Invoke(this, new EventArgs());
         }
         BackColor = ColorSave;
      }
   }
}
