﻿using IC_DB;
using IC_Tester_Controls;
using System;
using System.ComponentModel;
using System.Drawing;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using UrsComPortExtensions;
using static IC_Tester.TestWorker;

namespace IC_Tester
{

   /// <summary>
   /// Form zur Erfassung der Testschritte.
   /// </summary>
   public partial class IcTesterForm : Form
   {
      private class LogItem
      {
         public int StepNumber;
         public String StepName;
         public String State;
         public String ToolTipText;

         public LogItem(int StepNumber, String StepName, TestResult testResult)
         {
            this.StepNumber = StepNumber;
            this.StepName = StepName;
            State = testResult.GetDescription();
            ToolTipText = testResult.GetToolTipText();
         }
         public LogItem(int StepNumber, String StepName, String State)
         {
            this.StepNumber = StepNumber;
            this.StepName = StepName;
            this.State = State;
            ToolTipText = State;
         }


         public override string ToString()
         {
            return StepNumber.ToString("000") + " (" + StepName + ") " + State;
         }
      }

      private IcDefinition ic;
      private TestWorker testWorker = new TestWorker();
      private const string ComPortNameFilter = "";

      private bool enableF5 = false;
      private TestEditor? testEditorForm = null;
      Form owner;

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="IcTesterForm"/>-Klasse.
      /// </summary>
      /// <param name="ic">Das zu testende IC.</param>
      /// <param name="owner">Zum Ausblenden des aufrufenden Fensters.</param>
      public IcTesterForm(IcDefinition ic, Form owner)
      {
         InitializeComponent();
         this.ic = ic;
         Text = "IC-Tester: " + ic.Name;
         this.owner = owner;
      }

      private void IcTesterForm_Load(object sender, System.EventArgs e)
      {
         comPortSelector.UpdateList(ComPortNameFilter); // Arduino Nano hat diesen Typ
         pbCircuit.Image = ic.CircuitDiagram;

         testWorker.TestStepDone += testWorker_TestStepDone;
         testWorker.TestCompleted += testWorker_TestCompleted;
         testWorker.TestStepEnable += TestWorker_TestStepEnable;

         lbSteps.Items.Add("000: Start");
         lbSteps.Items.Add("000: PinMode");

         for (int i = 0; i < ic.TestSet.Count; i++)
         {
            lbSteps.Items.Add((i + 1).ToString("000") + ": " + ic.TestSet[i].Name);
         }
         lbSteps.Items.Add("000: Stopp");
         lbSteps.SelectedIndex = 0;
         owner.Visible = false;
         Show();
         testEditorForm = new TestEditor(ic, true);
         testEditorForm.StartPosition = FormStartPosition.Manual;
         testEditorForm.Location = new Point(this.Right + 5, this.Top);
         testEditorForm.Show();
         Activate();
      }

      private void comPortSelector_DeviceAdded(object sender, DeviceChangedEventArgs e)
      {
         comPortSelector.UpdateList();
      }

      private void comPortSelector_DeviceRemoved(object sender, DeviceChangedEventArgs e)
      {
         comPortSelector.UpdateList();
      }

      private void comPortSelector1_SelectionChanged(object sender, EventArgs e)
      {
         cmdStart.Enabled = comPortSelector.Selected != null;
      }

      private void cmdStart_Click(object sender, EventArgs e)
      {
         if (ArduinoPort.Connect(comPortSelector.Selected))
         {
            StatusLabel.Text = "Mit Arduino verbunden";
            StatusLabel.ForeColor = Color.Black;
            cmdStart.Enabled = false;
            cbSingleStep.Enabled = false;
            comPortSelector.Enabled = false;
            lbLog.Items.Clear();
            lbLog.Refresh();
            lbSteps.SelectedIndex = 0;
            testWorker.DoTest(ic, cbSingleStep.Checked);
            DisableFormCloseButton();
            testEditorForm?.ResetErrorMarks();
         }
         else
         {
            StatusLabel.Text = "Verbindung mit Arduino konnte nicht hergestellt werden";
            StatusLabel.ForeColor = Color.Red;
         }
      }

      private void IcTesterForm_FormClosing(object sender, System.Windows.Forms.FormClosingEventArgs e)
      {
         ArduinoPort.Disconnect();
         if (testEditorForm != null)
            testEditorForm.CloseForm();
         owner.Visible = true;
      }

      /// <summary>
      /// Liest das <see cref="DescriptionAttribute"/> des Enumerationswertes aus.
      /// </summary>
      /// <param name="value">Der Enumerationswerte, dessen <see cref="DescriptionAttribute"/> ermittelt werden soll.</param>
      /// <returns>Der beim Enumerationswert hintelegte Beschreibungstext oder der Name des Enumerationswerts.</returns>
      /// <typeparam name="T">Formaler Typ-Paramter.</typeparam>
      public static string GetEnumDescription<T>(T value) where T : Enum
      {
         FieldInfo? fi = value.GetType().GetField(value.ToString());

         if (fi != null)
         {
            DescriptionAttribute[] attributes = (DescriptionAttribute[])fi.GetCustomAttributes(typeof(DescriptionAttribute), false);
            if (attributes != null && attributes.Length > 0)
            {
               return attributes[0].Description;
            }
         }
         return value.ToString();
      }


      int aliveCount = 0;

      void testWorker_TestStepDone(object? sender, TestStepDoneEventArgs e)
      {
         if (e.State == TestStepDoneEventArgs.TestStepState.Start)
         {
            aliveCount = 0;
            StatusLabel.Text = "Testschritt " + e.Number + " (" + e.TestStepName + ") " + "Start";
            StatusLabel.ForeColor = Color.Red;
            lbLog.Items.Add(new LogItem(e.Number, e.TestStepName, "Start"));
            lbLog.TopIndex = lbLog.Items.Count - 1;
         }

         if (e.State == TestStepDoneEventArgs.TestStepState.Alive)
         {
            aliveCount++;
            String Dots = "";
            for (int i = 0; i < aliveCount; i++) { Dots += "."; }
            StatusLabel.Text = "Testschritt " + e.Number + " (" + e.TestStepName + ") " + " " + Dots;
            StatusLabel.ForeColor = Color.Red;
            lbLog.Items[lbLog.Items.Count - 1] = new LogItem(e.Number, e.TestStepName, Dots);
            lbLog.TopIndex = lbLog.Items.Count - 1;
         }

         if (e.State == TestStepDoneEventArgs.TestStepState.Finished)
         {
            StatusLabel.Text = "Testschritt " + e.Number + " (" + e.TestStepName + ") " + e.TestResult.GetDescription();
            StatusLabel.ForeColor = Color.Red;

            lbLog.Items[lbLog.Items.Count - 1] = new LogItem(e.Number, e.TestStepName, e.TestResult);

            lbLog.TopIndex = lbLog.Items.Count - 1;
            if (lbSteps.SelectedIndex < lbSteps.Items.Count - 2)
               lbSteps.SelectedIndex = lbSteps.SelectedIndex + 1;
            else
               lbSteps.SelectedIndex = -1;

            // TODO: Fehler sollten nicht anhand der Anzahl Fehler-Pins sondern am TestResult erkannt werden.
            if (e.ErrorPins.Count > 0 || e.TestResult == TestResult.InvalidI2CReadData)
            {
               if (testEditorForm != null)
                  testEditorForm.MarkErrors(e.Number, e.ErrorPins, e.TestResult.isI2cError());
            }
         }
      }

      void testWorker_TestCompleted(object? sender, TestCompletedEventArgs e)
      {
         ArduinoPort.Disconnect();
         StatusLabel.Text = "Test erledigt";
         StatusLabel.ForeColor = Color.Black;
         lbLog.Items.Add("Fertig");
         lbLog.TopIndex = lbLog.Items.Count - 1;
         cmdStart.Enabled = true;
         cbSingleStep.Enabled = true;
         lbSteps.SelectedIndex = 0;
         EnableFormCloseButton();
      }
      private void TestWorker_TestStepEnable(object? sender, EventArgs e)
      {
         cmdStep.Enabled = true;
         enableF5 = true;
         cmdTestAbort.Enabled = true;
      }

      private void cmdStep_Click(object sender, EventArgs e)
      {
         cmdStep.Enabled = false;
         enableF5 = false;
         cmdTestAbort.Enabled = false;
         testWorker.doNextStep = true;
      }

      private void cmdTestAbort_Click(object sender, EventArgs e)
      {
         cmdStep.Enabled = false;
         enableF5 = false;
         cmdTestAbort.Enabled = false;
         testWorker.cancel = true;
         lbSteps.SelectedIndex = 0;
      }

      private void lbSteps_DrawItem(object sender, DrawItemEventArgs e)
      {
         if (e.Index < 0)
            return;

         if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
            e.Graphics.FillRectangle(new SolidBrush(Color.Aqua),
                                     e.Bounds);
         else
            e.Graphics.FillRectangle(new SolidBrush(Color.White),
                                     e.Bounds);
         if (e.Font != null)
            e.Graphics.DrawString(lbSteps.Items[e.Index].ToString(),
                                  e.Font, new SolidBrush(Color.Black), e.Bounds, StringFormat.GenericDefault);
      }


      // https://social.msdn.microsoft.com/Forums/windows/en-US/32148556-2640-4d79-a7b6-f13f9ce3420a/how-to-enabledisable-the-close-buttonx-in-window-form-?forum=winforms
      // Letzter Eintrag


      [DllImport("user32")]
      private static extern int GetSystemMenu(IntPtr hWnd, bool bRevert);

      [DllImport("user32")]
      private static extern bool DeleteMenu(int hMenu, int uPosition, int uFlags);

      private int s_SystemMenuHandle = 0;

      // Deaktiviert den Form-Close-Button (X)
      private void DisableFormCloseButton()
      {
         s_SystemMenuHandle = GetSystemMenu(this.Handle, false);
         DeleteMenu(this.s_SystemMenuHandle, 6, 1024);
      }

      // Aktiviert den Form-Close-Button (X)
      private void EnableFormCloseButton()
      {
         this.s_SystemMenuHandle = GetSystemMenu(this.Handle, true);
         DeleteMenu(this.s_SystemMenuHandle, 6, 1024);
      }

      private void Form1_KeyDown(object sender, KeyEventArgs e)
      {
         if (e.Alt && e.KeyCode == Keys.F4)
         {
            e.Handled = true;
            return;
         }

         if (enableF5 && e.KeyCode == Keys.F5)
         {
            cmdStep_Click(this, new EventArgs());
            return;
         }
      }

      int CurrentItemIndex = -1;
      private void lbLog_MouseMove(object sender, MouseEventArgs e)
      {
         //Retrieve the index of the ListBox item at the current location. 
         CurrentItemIndex = lbLog.IndexFromPoint(e.Location);
         try
         {
            LogItem item = (LogItem)lbLog.Items[CurrentItemIndex];

            String text = item.ToolTipText;
            lbLogToolTip.SetToolTip(lbLog, text);
         }
         catch (Exception)
         {
         }
      }

      private void comPortSelector_ComPortFound(object sender, ComPortFoundEventArgs e)
      {
         e.ComPortDescriptor.DisplayName += " " + e.ComPortDescriptor.SerialNo;
         if (Main.ComPortSerialNo != "")
            e.AddDeviceToList = e.ComPortDescriptor.SerialNo.Contains(Main.ComPortSerialNo);
      }
   }
}