﻿using IC_DB;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;
using static IC_Tester.ArduinoPort;
using static IC_Tester.TestWorker.TestStepDoneEventArgs;

namespace IC_Tester
{
   /// <summary>
   /// Klasse zur Asynchronen Durführung der Tests.
   /// </summary>
   /// <remarks>Benutzt einen <see cref="BackgroundWorker"/>.</remarks>
   public class TestWorker
   {
      BackgroundWorker? worker = null;

      /// <summary>
      /// Argumente für das <see cref="TestStepDone"/>-Ereignis.
      /// </summary>
      public class TestStepDoneEventArgs
      {
         /// <summary>
         /// Status des Testschritts bei Rückmeldung über den Fortschritt.
         /// </summary>
         public enum TestStepState
         {
            /// <summary>Testschritt wurde gestartet.</summary>
            Start,
            /// <summary>Testschritt in Arbeit.</summary>
            Alive,
            /// <summary>Testschritt wurde beendet.</summary>
            Finished
         }

         /// <summary>Lfd. Testschritt-Nummer.</summary>
         public int Number;

         /// <summary>Die Bezeichnung des Testschritts.</summary>
         public String TestStepName;

         /// <summary>Das Ergebnis dieses Testschritts.</summary>
         public TestResult TestResult;

         /// <summary>Status des Testschritts bei Rückmeldung über den Fortschritt.</summary>
         public TestStepState State;

         /// <summary>Liste der Pins mit falschem Zustand.</summary>
         public List<int> ErrorPins = new List<int>();

         /// <summary>
         /// Initilaisiert eine neue Instanz der <see cref="TestStepDoneEventArgs"/>-Klasse.
         /// </summary>
         /// <param name="Number">Lfd. Testschritt-Nummer.</param>
         /// <param name="Name">Die Bezeichnung des Testschritts.</param>
         /// <param name="State">Status des Testschritts bei Rückmeldung über den Fortschritt.</param>
         /// <param name="TestResult">Das Ergebnis dieses Testschritts.</param>
         public TestStepDoneEventArgs(int Number, String Name, TestStepState State, TestResult TestResult)
         {
            this.Number = Number;
            this.TestStepName = Name;
            this.State = State;
            this.TestResult = TestResult;
         }

         /// <summary>
         /// Initilaisiert eine neue Instanz der <see cref="TestStepDoneEventArgs"/>-Klasse.
         /// </summary>
         /// <param name="Number">Lfd. Testschritt-Nummer.</param>
         /// <param name="Name">Die Bezeichnung des Testschritts.</param>
         /// <param name="State">Status des Testschritts bei Rückmeldung über den Fortschritt.</param>
         /// <param name="TestResult">Das Ergebnis dieses Testschritts.</param>
         /// <param name="ErrorPins">Liste der Pins mit falschem Zustand.</param>
         public TestStepDoneEventArgs(int Number, String Name, TestStepState State, TestResult TestResult, List<int> ErrorPins)
         {
            this.Number = Number;
            this.TestStepName = Name;
            this.State = State;
            this.TestResult = TestResult;
            this.ErrorPins = ErrorPins;
         }
      }

      /// <summary>
      /// Stellt die Methode dar, die das <see cref="TestStepDone"/>-Ereignis der <see cref="TestWorker"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="TestStepDoneEventArgs"/>-Objekt, das die Ereignisdaten enthält.</param>
      public delegate void TestStepDoneEventHandler(object? sender, TestStepDoneEventArgs e);
      /// <summary>
      /// Tritt ein, wenn ein Testschritt beendet wurde.
      /// </summary>
      public event TestStepDoneEventHandler? TestStepDone;

      /// <summary>
      /// Argumente für das <see cref="TestCompleted"/>-Ereignis.
      /// </summary>
      public class TestCompletedEventArgs
      {
         /// <summary><b>Exception</b>, wenn bei der Durchführung ein Fehler aufgetreten ist, ansonsten <b>null</b>.</summary>
         Exception? Exception;
         /// <summary>
         /// Initialisiert eine neue Instanz der <see cref="TestCompletedEventArgs"/>-Klasse.
         /// </summary>
         /// <param name="Exception"><b>Exception</b>, wenn bei der Durchführung ein Fehler aufgetreten ist, ansonsten <b>null</b>.</param>
         public TestCompletedEventArgs(Exception? Exception) { this.Exception = Exception; }
      }

      /// <summary>
      /// Stellt die Methode dar, die das <see cref="TestStepEnable"/>-Ereignis der <see cref="TestWorker"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="EventArgs"/>-Objekt, das die Ereignisdaten enthält.</param>
      public delegate void TestStepEnableEventHandler(object? sender, EventArgs e);
      /// <summary>
      /// Tritt ein, wenn ein Testschritt freigegeben werden soll.
      /// </summary>
      public event TestStepEnableEventHandler? TestStepEnable;


      private bool singleStep = false; // Einzelschrittausführung
      /// <summary>Bei Singlestep: Muss zum Fortfahren von außen vor jedem Schritt auf <b>true</b> gesetzt werden.
      /// </summary>
      public bool doNextStep = false; // Bei SingleStep true für Freigabe des nächsten Schritts

      /// <summary>Bei Singlestep: Bewirkt Abbruch des Tests-
      /// </summary>  
      public bool cancel = false; // Bei SingleStep true für Abbruch des Tests

      /// <summary>
      /// Stellt die Methode dar, die das <see cref="TestCompleted"/>-Ereignis der <see cref="TestWorker"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="TestCompletedEventArgs"/>-Objekt, das die Ereignisdaten enthält.</param>
      public delegate void TestCompletedEventHandler(object? sender, TestCompletedEventArgs e);
      /// <summary>
      /// Tritt ein, wenn alle Testschritte ausgeführt wurden oder ein Fehler bei der Kommunikation mit dem Arduino auftrat.
      /// </summary>
      public event TestCompletedEventHandler? TestCompleted;

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="TestWorker"/>-Klasse.
      /// </summary>
      public TestWorker() { }

      /// <summary>
      /// Führt die Testschritte asynchron aus.
      /// </summary>
      /// <param name="Ic">Das <see cref="IcDefinition"/>-Objekt mit den Testschritt-Definitionen.</param>
      /// <param name="SingleStep">Bei <b>true</b> wird vor jedem Testschritt auf Freigabe gewartet.</param>
      /// <exception cref="Exception"></exception>
      public void DoTest(IcDefinition Ic, bool SingleStep)
      {
         if (worker != null)
            throw new Exception("Worker is busy");

         singleStep = SingleStep;
         doNextStep = false;
         cancel = false;

         worker = new BackgroundWorker();
         worker.WorkerReportsProgress = true;
         worker.WorkerSupportsCancellation = true;
         worker.DoWork += worker_DoWork;
         worker.RunWorkerCompleted += worker_RunWorkerCompleted;
         worker.ProgressChanged += worker_ProgressChanged;
         worker.RunWorkerAsync(Ic);
      }

      // This event handler is where the time-consuming work is done.
      private void worker_DoWork(object? sender, DoWorkEventArgs e)
      {
         var arg = (IcDefinition?)e.Argument;
         if (arg == null)
         {
            e.Result = null;
            e.Cancel = true;
            return;
         }
         var Ic = (IcDefinition)arg;
         TestResult resultType = TestResult.TestStepOk;

         #region [ ---- Start-Paket ---- ]
         // -----------------------------------------------------------------------------
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Start", TestStepState.Start, TestResult.TestStepOk));
         ArduinoPort.BuildAndSendStartPacket(Ic);
         PacketGot rtc = ArduinoPort.ReadPacket();
         resultType = rtc.getTestResult();
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Start", TestStepState.Finished, resultType));
         if (resultType != TestResult.TestStepOk)
            return; // Kommunikationsfehler, Test abbrechen
         #endregion

         #region [ ---- PinMode ---- ]
         // -----------------------------------------------------------------------------
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "PinMode", TestStepState.Start, TestResult.TestStepOk));
         ArduinoPort.BuildAndSendPinModePacket(Ic);
         rtc = ArduinoPort.ReadPacket();
         resultType = rtc.getTestResult();
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "PinMode", TestStepState.Finished, resultType));

         if (resultType != TestResult.TestStepOk)
            return; // Kommunikationsfehler, Test abbrechen
         #endregion

         #region [ ---- Testschritte ---- ]
         // -----------------------------------------------------------------------------
         for (int testStepIndex = 0; testStepIndex < Ic.TestSet.Count; testStepIndex++)
         { // Über alle Testschritte

            if (singleStep)
            {
               worker?.ReportProgress(1); // Testschritt freigeben
               while (doNextStep == false && cancel == false)
                  Thread.Sleep(10); // Warten
               if (cancel)
               {
                  worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex, "Abbruch", TestStepState.Finished, TestResult.TestStepOk));
                  return;
               }
               doNextStep = false;
            }

            TestStep testStep = Ic.TestSet[testStepIndex];
            worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Start, TestResult.TestStepOk));


            ArduinoPort.BuildAndSendTestPacket(Ic, testStep);

            switch (testStep.StepType)
            {
               case TestStepType.IO:
               case TestStepType.I2CXmit:
                  checkIoTestStep(Ic, testStepIndex, testStep);
                  break;
               case TestStepType.I2CXmitAndRead:
                  checkI2cReadTestStep(Ic, testStepIndex, testStep);
                  break;
               default:
                  throw new Exception("Über alle Testschritte: Programmierfehler");
            }
         }
         #endregion

         #region [ ---- Start-Paket zum Abschluss (reset des MCP) ---- ]
         // -----------------------------------------------------------------------------
         // --- Start-Paket zum Abschluss (reset des MCP)
         /*
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Start, TestResult.ACK));

         ArduinoPort.SendStartPacket(Ic);
         rtc = ArduinoPort.ReadPacket();
         resultType = rtc.getTestResult();
         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Finished, resultType));
         */
         #endregion
      }

      #region [ ---- Auswertung der Testergebnisse ---- ]
      /// <summary>
      /// Prüft bei einem einfachen IO-Testschritt die zurück gelieferten Werte der Output-Pins gegen die Erwartung.
      /// </summary>
      /// <param name="Ic">Die IC-Definition.</param>
      /// <param name="testStepIndex">Die lfd. Nr. des Testschritts.</param>
      /// <param name="testStep">Die Testschritt-Daten.</param>
      private void checkIoTestStep(IcDefinition Ic, int testStepIndex, TestStep testStep)
      {
         PacketGot rtc = rtc = ArduinoPort.ReadPacket();
         TestResult resultType = rtc.getTestResult();

         var s = Ic.getPinsText();

         while (resultType == TestResult.TestStepOk) // ACK-Pakete = Alive-Kennung
         {
            worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Alive, TestResult.TestStepOk));
            rtc = ArduinoPort.ReadPacket();
            resultType = rtc.getTestResult();
         }

         if (resultType != TestResult.TestData)
         {
            worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Finished, resultType));
            return; // Kommunikationsfehler, Test abbrechen
         }

         resultType = TestResult.TestStepOk;
         // Empfangene Daten gegen Soll prüfen

         List<int> errorPins = new List<int>();

         for (int i = 0; i < Ic.NumberOfPins; i++)
         {
            if (Ic.Pins[i].Type == PinType.Output || Ic.Pins[i].Type == PinType.OpenCollector)
            {
               if (Ic.TestSet[testStepIndex].Pins[i] != LogicLevel.Ignore)
                  // in diesen Fällen muss der zurück gelieferte Wert dem im Testschritt entsprechen
                  if (rtc.Data[i] != (int)testStep.Pins[i])
                  {
                     // Fehler
                     errorPins.Add(i);
                  }
            }
         }
         if (errorPins.Count > 0)
            resultType = TestResult.TestStepError;

         // Testergebnis zurück melden, entweder TestResult.ACK oder TestResult.ACK
         worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Finished, resultType, errorPins));
      }

      /// <summary>
      /// Prüft bei einem einfachen IO-Testschritt die zurück gelieferten Werte der Output-Pins gegen die Erwartung.
      /// </summary>
      /// <param name="Ic">Die IC-Definition.</param>
      /// <param name="testStepIndex">Die lfd. Nr. des Testschritts.</param>
      /// <param name="testStep">Die Testschritt-Daten.</param>
      private void checkI2cReadTestStep(IcDefinition Ic, int testStepIndex, TestStep testStep)
      {
         PacketGot rtc = ArduinoPort.ReadPacket();
         TestResult resultType = rtc.getTestResult();

         while (resultType == TestResult.TestStepOk) // ACK-Pakete = Alive-Kennung
         {
            worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Alive, TestResult.TestStepOk));
            rtc = ArduinoPort.ReadPacket();
            resultType = rtc.getTestResult();
         }


         if (resultType != TestResult.TestI2cData)
         {
            worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Finished, resultType));
            return; // Kommunikationsfehler, Test abbrechen
         }

         //TODO: Erwartungswerte prüfen
         byte[] testData = testStep.I2cXmitData[testStep.I2cXmitData.Count - 1];
         for (int i = 1; i < testData.Length; i++)
         {
            if (testData[i] != rtc.Data[i - 1]) // In testData ist noch die Geräteadresse enthalten
            {
               resultType = TestResult.InvalidI2CReadData;
               rtc = ArduinoPort.ReadPacket(); // Es folgt noch das GPIO-Packet

               worker?.ReportProgress(0, new TestStepDoneEventArgs(testStepIndex + 1, testStep.Name, TestStepState.Finished, resultType));
               return;
            }
         }

         checkIoTestStep(Ic, testStepIndex, testStep);

      }
      #endregion

      #region [ ---- Rückmeldung über Events ---- ]
      private void worker_ProgressChanged(object? sender, ProgressChangedEventArgs e)
      {
         if (e.ProgressPercentage == 0)
         {
            if (e.UserState != null)
               TestStepDone?.Invoke(this, (TestStepDoneEventArgs)e.UserState);
            return;
         }
         if (e.ProgressPercentage == 1) // 1: Anfrage für Freigabe des nächsten Schritts.
         {
            TestStepEnable?.Invoke(this, EventArgs.Empty);
         }
      }

      // This event handler deals with the results of the background operation.
      private void worker_RunWorkerCompleted(object? sender, RunWorkerCompletedEventArgs e)
      {
         #region [ ---- Start-Paket zum Abschluss (reset des MCP) ---- ]
         // -----------------------------------------------------------------------------
         // --- Start-Paket zum Abschluss (reset des MCP)
         worker_ProgressChanged(this, new ProgressChangedEventArgs(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Start, TestResult.TestStepOk)));
         //        worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Start, TestResult.ACK));

         ArduinoPort.BuildAndSendStartPacket(null);
         var rtc = ArduinoPort.ReadPacket();
         var resultType = rtc.getTestResult();
         worker_ProgressChanged(this, new ProgressChangedEventArgs(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Finished, resultType)));
         //         worker?.ReportProgress(0, new TestStepDoneEventArgs(0, "Stopp", TestStepState.Finished, resultType));
         #endregion
         worker = null;
         TestCompleted?.Invoke(this, new TestCompletedEventArgs(e.Error));
      }
      #endregion
   }
}